
using System;
using System.Runtime.InteropServices;
using StreamInterfaceDriverNS;

namespace GIOcsWrapperNS
{
	/// <summary>
    /// GIOcsWrapper is based off of the StreamInterfaceDriver class
	/// and provides support for the general purpose IO (GIO) peripherals.
	/// </summary>
	public class GIOcsWrapper : StreamInterfaceDriver, IDisposable
	{
       
		// Track whether Dispose has been called.
		private bool disposed = false;

		public GIOcsWrapper() : base("GIO1:")
		{
		}

		/// <summary>
		/// Opens the GIOcsWrapper device.
		/// </summary>
		/// <returns>TRUE on success, otherwise FALSE </returns>
		public bool OpenDevice()
		{
			bool	fDeviceOpen = false;

			// Open the driver.
			if (base.Open())
			{
                fDeviceOpen = true;
			}
			return fDeviceOpen;
		}

		/// <summary>
		/// Closes the GIOcsWrapper device.
		/// </summary>
		public void CloseDevice()
		{
			base.Close();
		}

		/// <summary>
		/// Reads a value from a port.
		/// </summary>
		/// <param name="byPort">Port to read from.</param>
		/// <returns>Value read from port.</returns>
		public Byte ReadValue(Byte byPort)
		{
			Byte	byValue = 0x00;

			// Read a value from the port on the device.
			Byte[]	byDataPort = new Byte[] { byPort };
			UInt32	nDataInSize = (UInt32) byDataPort.GetLength(0);
			Byte[]	byDataValue = new Byte[1];
			UInt32	nDataOutSize = (UInt32) byDataValue.GetLength(0);
			UInt32	nBytesReturned = 0;

			base.DeviceIoControl(IOCTL_GIO_READ_BYTE, byDataPort, nDataInSize,
				byDataValue, nDataOutSize, out nBytesReturned);

			byValue = byDataValue[0];
			return byValue;
		}

        /// <summary>
        /// Reads a 16-bit value from a port.
        /// </summary>
        /// <param name="byPort">Port to read from.</param>
        /// <returns>Value read from port.</returns>
        public UInt16 Read16bitValue(Byte byPort)
        {
            UInt16 wValue = 0x0000;

            // Read a value from the port on the device.
            Byte[] byDataPort = new Byte[] { byPort };
            UInt32 nDataInSize = (UInt32)byDataPort.GetLength(0);
            Byte[] byDataValue = new Byte[2];
            UInt32 nDataOutSize = (UInt32)byDataValue.GetLength(0);
            UInt32 nBytesReturned = 0;

            base.DeviceIoControl(IOCTL_GIO_READ_BYTE, byDataPort, nDataInSize,
                byDataValue, nDataOutSize, out nBytesReturned);

            wValue = byDataValue[1];
            wValue <<= 8;
            wValue += byDataValue[0];
            return wValue;
        }

		/// <summary>
		/// Writes a value to a port.
		/// </summary>
		/// <param name="byPort">Port to write to.</param>
		/// <param name="byValue">Value to write.</param>
		public void WriteValue(Byte byPort, Byte byValue)
		{
			GIO_WRITE_VALUE		GIOWriteValue = new GIO_WRITE_VALUE();

			GIOWriteValue.byPortNumber = byPort;
			GIOWriteValue.byValue = byValue;

			// Write a value to the port in the device.
			UInt32	nDataSize = (UInt32) Marshal.SizeOf(GIOWriteValue);
			IntPtr	pGIOWriteValue = Memory.AllocHLocal(nDataSize);
			Marshal.StructureToPtr(GIOWriteValue, pGIOWriteValue, false);

			base.DeviceIoControl(IOCTL_GIO_WRITE_BYTE, pGIOWriteValue, nDataSize);
			Memory.FreeHLocal(pGIOWriteValue);
		}
		// Implement IDisposable.
		/// <summary>
		/// Dispose method for IDisposable.
		/// </summary>
		public void Dispose()
		{
			Dispose(true);

			// Instruct the GC to remove this object from the
			// Finalization queue to prevent finalization code
			// from executing a second time.
			GC.SuppressFinalize(this);
		}

		// Dispose(bool disposing) executes in two distinct scenarios.
		// If disposing equals true, the method has been called directly
		// or indirectly by a user's code. Managed and unmanaged resources
		// can be disposed.
		// If disposing equals false, the method has been called by the
		// runtime from inside the finalizer and you should not reference
		// other objects. Only unmanaged resources can be disposed.
		protected virtual void Dispose(bool disposing)
		{
			// Check to see if Dispose has already been called.
			if (!this.disposed)
			{
				// if disposing equals true, dispose all managed
				// and unmanged resources.
				if (disposing)
				{
					// Dispose managed resources.
				}

				// Release unmanaged resources. If disposing is false,
				// only the following code is executed.
				base.Close();
			}
		}

		// This finalizer will run only if the Dispose method
		// does not get called.
        ~GIOcsWrapper()
		{
			// Do not re-create Dispose clean-up code here.
			// Calling Dispose(false) is optimal.
			Dispose(false);
		}

		// Structure to use when writing a value to an
		// I/O point or port using IOCTL_GIO_WRITE_BIT
		// and IOCTL_GIO_WRITE_BYTE.
		private struct GIO_WRITE_VALUE
		{
			//	Port number to access.
			public	Byte	byPortNumber;

			// Value to write to port.
			public	Byte	byValue;
		};

		// Define the type of device.
		// This parameter can be no bigger than a WORD value. The values used by
		// Microsoft are in the range of 0-32767 and the values between
		// 32768-65535 are reserved for use by OEMs and IHVs.
		private const UInt32	GIO_DEV_TYPE	= 32769;

        // Create  unique system I/O control codes (IOCTL).
        // The function codes 0-2047 are reserved for Microsoft and 2048-4095 are
        // reserved for OEMs and IHVs.  The function codes can be no larger than 4095.


        // ***********************************************************
        //
        //	IOCTL_GIO_READ_BYTE
        //
        //	Device IO control routine to get the state of the I/O pins
        //	in the specified port.
        //
        //	dwCode			IOCTL_GIO_READ_BYTE
        //	pBufIn			Pointer to a BYTE containing the port number
        //					to read.
        //
        //					Valid Port Numbers:
        //						SEE "Port access definitions" above
        //
        //	dwLenIn			Size of pBufIn, should be sizeof(BYTE).
        //	pBufOut			Pointer to a BYTE to receive the state
        //					of the I/O port.
        //	dwLenOut		Size of pBufOut, should be sizeof(BYTE).
        //	pdwActualOut	Points to a DWORD to return the number
        //					of bytes returned in pBufOut if no error.
        //
		private const UInt32 IOCTL_GIO_READ_BYTE = 
			((GIO_DEV_TYPE) << 16) | ((FILE_ANY_ACCESS) << 14) | ((0x0801) << 2) | (METHOD_NEITHER);

        // ***********************************************************
        //
        //	IOCTL_GIO_WRITE_BYTE
        //
        //	Device IO control routine to set the state of the I/O pins
        //	in the specified port.
        //
        //	dwCode			IOCTL_GIO_WRITE_BYTE
        //	pBufIn			Pointer to a GIO_WRITE_VALUE structure.
        //
        //					GIO_WRITE_VALUE structure
        //					wPortNumber	= The port number to write value to.
        //						Valid Port Numbers:
        //							SEE "Port access definitions" above

        //					byValue		= Value to write to the port.
        //					
        //	dwLenIn			Size of pBufIn, should be sizeof(GIO_WRITE_VALUE).
        //	pBufOut			Ignored.
        //	dwLenOut		Ignored.
        //	pdwActualOut	Ignored.
        //
		private const UInt32 IOCTL_GIO_WRITE_BYTE = 
			((GIO_DEV_TYPE) << 16) | ((FILE_ANY_ACCESS) << 14) | ((0x0802) << 2) | (METHOD_NEITHER);


		//
		// Define the method codes for how buffers are passed for I/O and FS controls
		//

		private const UInt32	METHOD_BUFFERED		= 0;
		private const UInt32	METHOD_IN_DIRECT	= 1;
		private const UInt32	METHOD_OUT_DIRECT	= 2;
		private const UInt32	METHOD_NEITHER		= 3;

		//
		// Define the access check value for any access
		//
		//
		// The FILE_READ_ACCESS and FILE_WRITE_ACCESS constants are also defined in
		// ntioapi.h as FILE_READ_DATA and FILE_WRITE_DATA. The values for these
		// constants *MUST* always be in sync.
		//

		private const UInt32	FILE_ANY_ACCESS		= 0;
		private const UInt32	FILE_READ_ACCESS	= 0x0001;	// file & pipe
		private const UInt32	FILE_WRITE_ACCESS	= 0x0002;	// file & pipe
	}

	/// <summary>
	/// Memory class is used in the process of converting a 
	/// structure to a pointer for marshaling.
	/// </summary>
	internal class Memory
	{
		// Allocates a block of memory using LocalAlloc.
		public static IntPtr AllocHLocal(UInt32 nSize)
		{
			return LocalAlloc(LPTR, nSize);
		}

		// Frees memory allocated by AllocHLocal.
		public static void FreeHLocal(IntPtr hLocal)
		{
			if (!hLocal.Equals(IntPtr.Zero))
			{
				LocalFree(hLocal);
			}
		}

		[DllImport("coredll.dll", SetLastError=true)]
		internal static extern IntPtr LocalAlloc(UInt32 uFlags, UInt32 uBytes);

		[DllImport("coredll.dll", SetLastError=true)]
		internal static extern IntPtr LocalFree(IntPtr hMem);

		private const UInt32 LMEM_FIXED		= 0;
		private const UInt32 LMEM_MOVEABLE	= 2;
		private const UInt32 LMEM_ZEROINIT	= 0x40;
		private const UInt32 LPTR			= (LMEM_FIXED | LMEM_ZEROINIT);
	}
}
