// 
//
using System;
using System.IO;
using System.Runtime.InteropServices;

namespace StreamInterfaceDriverNS
{
	/// <summary>
	/// StreamInterfaceDriver is a base class for stream interface drivers.
	/// It handles the P/Invoke to access the functions of a stream
	/// interface driver.
	/// </summary>
	public abstract class StreamInterfaceDriver
	{
		IntPtr	m_hPort = (IntPtr) Win32File.INVALID_HANDLE_VALUE;
		string	m_portName = null;

		/// <summary>
		/// StreamInterfaceDriver Constructor.
		/// </summary>
		/// <param name="portName">The name of the device "COM1:".</param>
		public StreamInterfaceDriver(
			string portName)
		{
			m_portName = portName;
		}

		/// <summary>
		/// Opens the device.
		/// </summary>
		/// <returns></returns>
		protected bool Open()
		{
			m_hPort = Win32File.CreateFile(	m_portName,
				Win32File.GENERIC_READ | Win32File.GENERIC_WRITE,
				0,
				IntPtr.Zero,
				Win32File.OPEN_EXISTING,
				0,
				IntPtr.Zero);
			if (m_hPort == (IntPtr) Win32File.INVALID_HANDLE_VALUE)
			{
				if (Marshal.GetLastWin32Error() == Win32File.ERROR_ACCESS_DENIED)
				{
					return false;
				}
			}

			return true;
		}

		/// <summary>
		/// Closes the device.
		/// </summary>
		protected void Close()
		{
			if ((IntPtr) Win32File.INVALID_HANDLE_VALUE != m_hPort)
			{
				Win32File.CloseHandle(m_hPort);
				m_hPort = (IntPtr) Win32File.INVALID_HANDLE_VALUE;
			}
		}

		/// <summary>
		/// Performs an IOCTL function.
		/// </summary>
		/// <param name="dwIoControlCode">IOCTL function to perform.</param>
		/// <param name="lpInBuffer">Input data to the function.</param>
		/// <param name="nInBufferSize">Size of the input data.</param>
		/// <returns></returns>
		protected bool DeviceIoControl(
			UInt32		dwIoControlCode,
			Byte[]		lpInBuffer,
			UInt32		nInBufferSize)
		{
			UInt32 lpBytesReturned = 0;
			return Win32File.DeviceIoControl(m_hPort, dwIoControlCode, lpInBuffer,
				nInBufferSize, null, 0, ref lpBytesReturned, IntPtr.Zero);
		}

		/// <summary>
		/// Performs an IOCTL function.
		/// </summary>
		/// <param name="dwIoControlCode">IOCTL function to perform.</param>
		/// <param name="lpInBuffer">Input data to the function.</param>
		/// <param name="nInBufferSize">Size of the input data.</param>
		/// <returns></returns>
		protected bool DeviceIoControl(
			UInt32		dwIoControlCode,
			IntPtr		lpInBuffer,
			UInt32		nInBufferSize)
		{
			UInt32 lpBytesReturned = 0;
			return Win32File.DeviceIoControl(m_hPort, dwIoControlCode, lpInBuffer,
				nInBufferSize, null, 0, ref lpBytesReturned, IntPtr.Zero);
		}

		/// <summary>
		/// Performs an IOCTL function.
		/// </summary>
		/// <param name="dwIoControlCode">IOCTL function to perform.</param>
		/// <param name="lpInBuffer">Input data to the function.</param>
		/// <param name="nInBufferSize">Size of the input data.</param>
		/// <param name="lpOutBuffer">Data output from the function.</param>
		/// <param name="nOutBufferSize">Size of the output data buffer.</param>
		/// <param name="lpBytesReturned">Actual number of bytes placed in the output data buffer.</param>
		/// <returns></returns>
		protected bool DeviceIoControl(
			UInt32		dwIoControlCode,
			Byte[]		lpInBuffer,
			UInt32		nInBufferSize,
			Byte[]		lpOutBuffer,
			UInt32		nOutBufferSize,
			out UInt32	lpBytesReturned)
		{
			lpBytesReturned = 0;
			return Win32File.DeviceIoControl(m_hPort, dwIoControlCode, lpInBuffer,
				nInBufferSize, lpOutBuffer, nOutBufferSize, ref lpBytesReturned, IntPtr.Zero);
		}
	}

	internal class Win32File
	{
		[DllImport("coredll.dll", SetLastError=true)]
		internal static extern IntPtr CreateFile(String lpFileName, UInt32 dwDesiredAccess, UInt32 dwShareMode,
			IntPtr lpSecurityAttributes, UInt32 dwCreationDisposition, UInt32 dwFlagsAndAttributes,
			IntPtr hTemplateFile);

		[DllImport("coredll.dll")]
		internal static extern Boolean CloseHandle(IntPtr hObject);

		[DllImport("coredll.dll")]
		internal static extern Boolean DeviceIoControl(IntPtr hDevice, UInt32 dwIoControlCode,
			Byte[] lpInBuffer, UInt32 nInBufferSize, Byte[] lpOutBuffer, UInt32 nOutBufferSize,
			ref UInt32 lpBytesReturned, IntPtr lpOverlapped);

		[DllImport("coredll.dll")]
		internal static extern Boolean DeviceIoControl(IntPtr hDevice, UInt32 dwIoControlCode,
			IntPtr lpInBuffer, UInt32 nInBufferSize, Byte[] lpOutBuffer, UInt32 nOutBufferSize,
			ref UInt32 lpBytesReturned, IntPtr lpOverlapped);

		// Constants for errors:
		internal const UInt32 ERROR_FILE_NOT_FOUND = 2;
		internal const UInt32 ERROR_INVALID_NAME = 123;
		internal const UInt32 ERROR_ACCESS_DENIED = 5;
		internal const UInt32 ERROR_IO_PENDING = 997;
		internal const UInt32 ERROR_IO_INCOMPLETE = 996;

		// Constants for return value:
		internal const Int32 INVALID_HANDLE_VALUE = -1;

		// Constants for dwFlagsAndAttributes:
		internal const UInt32 FILE_FLAG_OVERLAPPED = 0x40000000;

		// Constants for dwCreationDisposition:
		internal const UInt32 OPEN_EXISTING = 3;

		// Constants for dwDesiredAccess:
		internal const UInt32 GENERIC_READ = 0x80000000;
		internal const UInt32 GENERIC_WRITE = 0x40000000;
	}
}
